//==================================
// LOADPROF - Matt Pietrek 1994
// FILE: LOADPROF.C
//==================================
#include <windows.h>
#include <windowsx.h>
#include <toolhelp.h>
#include <commdlg.h>
#include <string.h>
#include <stdio.h>
#include "loadprof.h"
#include "ldevent.h"
#include "newexe.h"

VOID    FAR PASCAL InitTask(void);      // Prototypes
void    LoadProf_IntHandler(void);
DWORD   GetMillisecondCount(void);
void    AddLoadProfEvent( LPLOADPROF_EVENT );
void    StopWatching(void);
BOOL    GetProgramName(HWND hWndOwner, LPSTR szFile, unsigned nFileBuffSize);
BOOL    IsSegmentPreload(LPCSTR lpszModule, WORD segNumber);
void    WriteReportToFile(HWND hWndOwner);
void    ScanForDuplicateFixups(void);
DWORD   HasDuplicateFixups(LPSTR szModuleName); // In DUPFIXUP.C

HWND HWndEventsLb;          // HWND's of commonly used dialog controls
HWND HWndDlg;

BOOL WaitingForLoad = FALSE;
BOOL WaitingForGetMessage = FALSE;
HHOOK hHook=0;                      // WH_GETMESSAGE hook
HANDLE HInstance;                   // hInstance of this program
HTASK hTask;                        // hTask of WinExec()'ed task

FARPROC NotifyCallbackThunk;        // MakeProcInstance() thunks.  If
FARPROC GetMsgProcThunk;            // Borland's _loadds worked, we
FARPROC DialogBoxThunk;             // wouldn't need these

LPLOADPROF_EVENT LoadProfEvents = 0;    // LoadProfEvents array info
WORD cLoadProfEvents = 0;
WORD maxLoadProfEvents = 0;

LPBYTE LpfnInitTaskReturn = 0;          // InitTask breakpoint stuff
BYTE   OriginalInitTaskOpcode;
BYTE   OriginalInitTaskReturnOpcode;

//
// Minimal WH_GETMESSAGE hook procedure.  Once installed, it simply
// looks for the first GetMessage (or PeekMessage) call from the
// application that we WinExec'ed.
//
LRESULT CALLBACK _export GetMsgProc(int code, WPARAM wParam, LPARAM lParam)
{
    if ( hTask == GetCurrentTask() )        // Is it the right task?
        if ( WaitingForGetMessage == TRUE ) // Are we looking for one?
            StopWatching();

    return CallNextHookEx(hHook, code, wParam, lParam);
}

//
// TOOLHELP NotifyRegister() callback function.
//
BOOL CALLBACK _export NotifyCallback(WORD wID, DWORD dwData)
{
    LOADPROF_EVENT event;

    //
    // Most notifications get no further than here.
    //
    if ( (WaitingForLoad == FALSE) && (WaitingForGetMessage==FALSE) )
        return FALSE;

    //
    // If we get here, we're somewhere in the process of loading the
    // specified program.
    //
    if ( wID == NFY_LOADSEG )
    {
        LPCSTR lpszModuleName = ((NFYLOADSEG FAR *)dwData)->lpstrModuleName;
        WORD wSegNum = ((NFYLOADSEG FAR *)dwData)->wSegNum;
        event.dwMilliseconds = GetMillisecondCount();
        event.type = LOADSEG_EVENT;
        event.x.loadSeg.wSegNum = wSegNum;
        event.x.loadSeg.aSzModule = AddAtom( lpszModuleName );
        event.x.loadSeg.fPreload = IsSegmentPreload(lpszModuleName, wSegNum);
        AddLoadProfEvent( &event );
    }
    else if ( wID == NFY_STARTTASK )
    {
        BYTE bpOpcode = 0xCC;
        char szFilename[260];

        if ( WaitingForLoad == FALSE )
            return FALSE;

        // Fill in and add the STARTTASK_EVENT to the list
        event.dwMilliseconds = GetMillisecondCount();
        GetModuleFileName(*(LPWORD)MAKELP(GetCurrentTask(), 0x1E),
                            szFilename, sizeof(szFilename));
        event.x.startTask = AddAtom( szFilename );
        event.type = STARTTASK_EVENT;
        AddLoadProfEvent( &event );

        WaitingForLoad = FALSE;
        WaitingForGetMessage = TRUE;

        // After we get the NFY_STARTTASK notifications, all the preload
        // segments are in memory.  However, other segments may be
        // loaded before the task enters its message loop.  Therefore
        // we set up a WH_GETMESSAGE hook to look for the first call.
        hTask = GetCurrentTask();
        hHook = SetWindowsHookEx(WH_GETMESSAGE, (HOOKPROC)GetMsgProcThunk,
                                 HInstance, 0);

        // Register our interrupt handler, then set the breakpoint on
        // InitTask()
        InterruptRegister( 0,
                    MakeProcInstance((FARPROC)LoadProf_IntHandler,HInstance));
        MemoryRead( SELECTOROF(InitTask), OFFSETOF(InitTask),
                    &OriginalInitTaskOpcode, sizeof(BYTE) );
        MemoryWrite( SELECTOROF(InitTask), OFFSETOF(InitTask),
                     &bpOpcode, sizeof(BYTE) );
    }
    else if ( wID == NFY_STARTDLL )
    {
        char filename[260];

        event.dwMilliseconds = GetMillisecondCount();
        event.type = STARTDLL_EVENT;

        GetModuleFileName( ((NFYSTARTDLL FAR *)dwData)->hModule,
                            filename, sizeof(filename) );
        event.x.startDll = AddAtom( filename );

        AddLoadProfEvent( &event );
    }
    else if ( (wID == NFY_EXITTASK) && (GetCurrentTask() == hTask) )
    {
        // In case the app went all the way through without calling
        // GetMessage or PeekMessage, fake the GETMESSAGE_EVENT
        StopWatching();
    }

    return FALSE;   // Chain on to next handler (Unlike certain IDEs!)
}

//
// Breakpoint hander called by LoadProf_IntHandler in INT.ASM
//
BOOL FAR PASCAL Int3Handler(LPBYTE csip, LPDWORD sssp)
{
    BYTE bpOpcode = 0xCC;   // The INT 3 opcode

    if ( GetCurrentTask() != hTask )    // Wrong task!  Let TOOLHELP
        return FALSE;                   // keep searching for a handler

    csip--;     // Back up 1 byte to point at faulting address;

    if ( csip == (LPBYTE)InitTask )
    {
        // Retrieve the address InitTask will be returning to
        LpfnInitTaskReturn = (LPBYTE) *sssp;

        // Now read in the initial BYTE at that address
        MemoryRead(SELECTOROF(LpfnInitTaskReturn),OFFSETOF(LpfnInitTaskReturn),
                    &OriginalInitTaskReturnOpcode, sizeof(BYTE));

        // Now write out a 2nd breakpoint to that address
        MemoryWrite(SELECTOROF(LpfnInitTaskReturn),OFFSETOF(LpfnInitTaskReturn),
                    &bpOpcode, sizeof(BYTE));

        // put the original InitTask opcode back
        MemoryWrite( SELECTOROF(csip), OFFSETOF(csip),
                        &OriginalInitTaskOpcode, sizeof(BYTE) );

        return TRUE;    // Restart the instruction
    }
    else if ( csip == LpfnInitTaskReturn )  // InitTaskReturn BP has been hit
    {
        LOADPROF_EVENT event;

        // Save off the event information
        event.dwMilliseconds = GetMillisecondCount();
        event.type = POSTDLL_EVENT;
        AddLoadProfEvent( &event );

        // put the original InitTaskReturn opcode back
        MemoryWrite( SELECTOROF(csip), OFFSETOF(csip),
                        &OriginalInitTaskReturnOpcode, sizeof(BYTE) );

        InterruptUnRegister( 0 );   // Don't need this anymore

        return TRUE;    // Restart the instruction
    }

    return FALSE;       // Let TOOLHELP keep searching for a handler
}

//
// Clean out any data from previous runs, then install the
// NotifyRegister handler.
//
void PrepareToLoad(void)
{
    LPLOADPROF_EVENT lpEvent = LoadProfEvents;
    unsigned i;

    // Clear out any atoms we have from previous runs
    for ( i = 0; i < cLoadProfEvents; i++, lpEvent++ )
    {
        if ( lpEvent->type == STARTDLL_EVENT )
            DeleteAtom( lpEvent->x.startDll );
        else if ( lpEvent->type == STARTTASK_EVENT )
            DeleteAtom( lpEvent->x.startTask );
        else if ( lpEvent->type == LOADSEG_EVENT )
            DeleteAtom( lpEvent->x.loadSeg.aSzModule );
    }

    cLoadProfEvents = 0;
    WaitingForLoad = TRUE;

    // Clear the listbox
    SendMessage(HWndEventsLb, LB_RESETCONTENT, 0, 0);

    // Install the callback to see segment load & task starts
    NotifyRegister(0, (LPFNNOTIFYCALLBACK)NotifyCallbackThunk,
                    NF_NORMAL);
}

//
// Iterates through all the segments that were loaded as a result of the
// WinExec(). Adds relevevant data to one or both listboxes.
//
void GenerateReport(HWND hWndDlg)
{
    HCURSOR oldCursor = 0;
    char buffer[384];
    char filename[260];
    LPLOADPROF_EVENT current, nextBig, endEvent, startTaskEvent, postDllEvent;
    BOOL fExit;

    //
    // This may take awhile, so put up the hourglass cursor and
    // disable listbox updating.
    //
    oldCursor = SetCursor( LoadCursor(0, IDC_WAIT) );
    SendMessage(HWndEventsLb, WM_SETREDRAW, FALSE, NULL);

    nextBig = LoadProfEvents;
    endEvent = &LoadProfEvents[cLoadProfEvents];

    SendMessage(HWndEventsLb,LB_ADDSTRING,0,
            (LPARAM)(LPSTR)"***** PRELOAD segment loading complete"
                           " - Calling DLL LibMain's *****");

    // Scan past preload segment loading
    while ( nextBig->type != STARTTASK_EVENT )
        nextBig++;

    // fill in the "Seg Load:" field
    wsprintf(buffer, "%u ms",
            nextBig->dwMilliseconds - LoadProfEvents[0].dwMilliseconds);
    SetDlgItemText(hWndDlg, IDC_STATIC_SEG_LOAD, buffer);

    startTaskEvent = nextBig;

    //
    // DLL loading - Scan till either the first loaded DLL, or the
    // return from InitTask (i.e., POSTDLL_EVENT)
    //
    while ( (nextBig->type != STARTDLL_EVENT)
            && (nextBig->type != POSTDLL_EVENT)
            && (nextBig < endEvent) )
        nextBig++;

    fExit = FALSE;
    while ( nextBig < endEvent )
    {
        // Remember the current index of the end of the listbox.  We're
        // going to add segment load events (possibly), but we want the
        // module load event to be inserted before the segment loads
        WORD modLoadInsertionIndex =
            (WORD)SendMessage(HWndEventsLb, LB_GETCOUNT, 0, 0);

        current = nextBig;  // Base us at the current NFY_STARTDLL
        nextBig++;          // move scan pointer to next event

        // Scan to the next STARTDLL_EVENT or POSTDLL_EVENT
        while (    (nextBig->type != STARTDLL_EVENT)
                && (nextBig->type != POSTDLL_EVENT)
                && (nextBig < endEvent) )
        {
            // If it's an NFY_LOADSEG event, the segment should have
            // been marked PRELOAD
            if ( nextBig->type == LOADSEG_EVENT )
            {
                GetAtomName( nextBig->x.loadSeg.aSzModule,
                             filename, sizeof(filename) );

                // Find out if the segment was demand loaded, rather than
                // preloaded.  If so, rat on it.
                if ( ! nextBig->x.loadSeg.fPreload )
                {
                    wsprintf(buffer, "    %s segment %Xh should be preload",
                            (LPSTR)filename, nextBig->x.loadSeg.wSegNum);
                    SendMessage(HWndEventsLb,LB_ADDSTRING,0,
                                (LPARAM)(LPSTR)buffer);
                }
            }

            nextBig++;
        }

        // Have we hit the end of the event array?  Set flag so that we'll
        // break out of the loop after reporting this last STARTDLL_EVENT
        if ( nextBig == endEvent )
        {
            nextBig--;
            fExit = TRUE;
        }

        if ( current->type == STARTDLL_EVENT )
        {
            // Report a DLL loading
            GetAtomName( current->x.startDll, filename, sizeof(filename) );
            wsprintf(buffer, "DLL load: %s %u ms", (LPSTR)filename,
                    nextBig->dwMilliseconds - current->dwMilliseconds);
            SendMessage(HWndEventsLb,LB_INSERTSTRING,
                        modLoadInsertionIndex,(LPARAM)(LPSTR)buffer);
        }
        else if ( current->type == POSTDLL_EVENT )
        {
            // Report the end of DLL loading
            postDllEvent = current;
            wsprintf(buffer, "***** DLL loading complete *****");
            SendMessage(HWndEventsLb,LB_INSERTSTRING,
                        modLoadInsertionIndex,(LPARAM)(LPSTR)buffer);

            // Update the "DLL Init:" field
            wsprintf(buffer, "%u ms",
                postDllEvent->dwMilliseconds-startTaskEvent->dwMilliseconds);
            SetDlgItemText(hWndDlg, IDC_STATIC_DLL_INIT, buffer);
        }

        if ( fExit )
            break;
    }

    wsprintf(buffer, "%u ms",
            nextBig->dwMilliseconds - postDllEvent->dwMilliseconds);
    SetDlgItemText(hWndDlg, IDC_STATIC_APP_INIT, buffer);

    wsprintf(buffer, "%u ms",
            nextBig->dwMilliseconds - LoadProfEvents[0].dwMilliseconds);
    SetDlgItemText(hWndDlg, IDC_STATIC_TOTAL_TIME, buffer);

    //
    // Restore cursor and listbox updating to normal.
    //
    SendMessage(HWndEventsLb, WM_SETREDRAW, TRUE, NULL);
    SetCursor(oldCursor);

    EnableWindow( GetDlgItem(hWndDlg, IDC_BUTTON_WRITE_REPORT), TRUE );
    EnableWindow( GetDlgItem(hWndDlg, IDC_BUTTON_SCAN), TRUE );
}

//
// Handle the dialog's WM_COMMAND messages
//
void Handle_WM_COMMAND(HWND hWndDlg, WPARAM wParam, LPARAM lParam)
{
    char buffer[260];

    switch ( wParam )
    {
        case IDC_BUTTON_LOAD:
            PrepareToLoad();    // Set global vars, etc...

            // Get the command line from the edit control and WinExec() it
            GetDlgItemText(hWndDlg, IDC_EDIT_PROGRAM, buffer, sizeof(buffer));
            WriteProfileString( "LOADPROF", "CmdLine", buffer );
            if ( WinExec(buffer, SW_SHOW) < 0x20 )
            {
                NotifyUnRegister(0);
                MessageBox(hWndDlg, "Couldn't WinExec() program", 0, MB_OK);
            }
            return;

        case IDC_BUTTON_BROWSE:
            if ( GetProgramName(hWndDlg, buffer, sizeof(buffer)) )
                SetWindowText( GetDlgItem(hWndDlg, IDC_EDIT_PROGRAM), buffer);
            break;

        case IDC_BUTTON_SCAN:
            ScanForDuplicateFixups();
            break;

        case IDC_BUTTON_WRITE_REPORT:
            WriteReportToFile(hWndDlg);
            break;

        case IDC_BUTTON_EXIT:
            EndDialog(hWndDlg, 0);
            break;

        case IDM_PROGRAM_YIELDED:
            // The program has finally called GetMessage (or PeekMessage).
            // Undo the WH_GETMESSAGE hook, then generate the report
            NotifyUnRegister(0);
            UnhookWindowsHookEx(hHook);
            hHook = hTask = 0;
            GenerateReport(hWndDlg);
            return;
    }
    return;
}

void Handle_WM_INITDIALOG(HWND hWndDlg)
{
    char buffer[260];

    HWndDlg = hWndDlg;  // Save away in a global variable
                        // We'll need it later to do a PostMessage()

    // Disable the two buttons that don't do anything till there's a report
    EnableWindow( GetDlgItem(hWndDlg, IDC_BUTTON_WRITE_REPORT), FALSE );
    EnableWindow( GetDlgItem(hWndDlg, IDC_BUTTON_SCAN), FALSE );

    // Get the HWND's of the commonly modified controls
    HWndEventsLb = GetDlgItem(hWndDlg, IDC_LISTBOX_EVENTS);

    // Retrieve the last command line, and use it as the default
    GetProfileString("LOADPROF", "CmdLine", "", buffer,sizeof(buffer));
    SetDlgItemText( hWndDlg, IDC_EDIT_PROGRAM, buffer );

    SetFocus( GetDlgItem(hWndDlg, IDC_EDIT_PROGRAM) );
}

//
// Dialog proc for the main dialog
//
BOOL CALLBACK _export LoadProfDlgProc(HWND hWndDlg, UINT msg,
                                     WPARAM wParam, LPARAM lParam)
{
    switch ( msg )
    {
        case WM_COMMAND:
            Handle_WM_COMMAND(hWndDlg, wParam, lParam); return TRUE;
        case WM_INITDIALOG:
            Handle_WM_INITDIALOG(hWndDlg); return TRUE;
        case WM_CLOSE:
            EndDialog(hWndDlg, 0); return FALSE;
    }
    return FALSE;
}

BOOL GlobalInitialize(HANDLE hInstance)
{
    HInstance = hInstance;

    //
    // Get the required MakeProcInstance() thunks.  We wouldn't need
    // these if we could rely on _loadds working properly.
    //
    NotifyCallbackThunk = MakeProcInstance(NotifyCallback, hInstance);
    GetMsgProcThunk = MakeProcInstance((FARPROC)GetMsgProc, hInstance);
    DialogBoxThunk = MakeProcInstance(LoadProfDlgProc, hInstance);

    // Set up the information for maintaining the LoadProfEvents array
    LoadProfEvents = (LPLOADPROF_EVENT)GlobalAllocPtr(GMEM_MOVEABLE, 65536L);
    if ( !LoadProfEvents )
    {
        MessageBox(0,"Unable to allocated needed memory - exiting", 0, MB_OK);
        return FALSE;
    }
    maxLoadProfEvents = (WORD)(65536L / sizeof(LOADPROF_EVENT));

    return TRUE;
}

void GlobalShutdown(void)
{
    // Undo the TOOLHELP and hook callbacks and free the
    // MakeProcInstance() thunks.
    if ( hHook )
        UnhookWindowsHookEx(hHook);
    FreeProcInstance(NotifyCallbackThunk);
    FreeProcInstance(GetMsgProcThunk);
    FreeProcInstance(DialogBoxThunk);

    GlobalFreePtr( LoadProfEvents );
}

int PASCAL WinMain( HANDLE hInstance, HANDLE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow )
{
    if ( hPrevInstance )
        return 0;

    if ( !GlobalInitialize(hInstance) )
        return 0;

    DialogBox(hInstance, "LoadProfDlg", 0, (DLGPROC)DialogBoxThunk);

    GlobalShutdown();

    return 0;
}

// Call TOOLHELP to get the current number of MS for this virtual machine
DWORD GetMillisecondCount(void)
{
    TIMERINFO ti;
    ti.dwSize = sizeof(ti);
    TimerCount(&ti);
    return ti.dwmsThisVM;
}

void AddLoadProfEvent( LPLOADPROF_EVENT event)
{
    if ( cLoadProfEvents >= maxLoadProfEvents )
        return;

    LoadProfEvents[ cLoadProfEvents++ ] = *event; // Let compiler do the copy
}

// Add the last event (the GETMESSAGE_EVENT) to the LoadProvEvents array,
// then post a message to LOADPROF's window, indicating it's done recording
void StopWatching(void)
{
    LOADPROF_EVENT event;

    event.dwMilliseconds = GetMillisecondCount();
    event.type = GETMESSAGE_EVENT;
    AddLoadProfEvent( &event );

    WaitingForGetMessage = FALSE;   // No more segment loads please!
    PostMessage(HWndDlg, WM_COMMAND, IDM_PROGRAM_YIELDED, 0);
}

// Return status of the PRELOAD bit in the module's segment table
BOOL IsSegmentPreload(LPCSTR lpszModule, WORD segNumber)
{
    HMODULE_SEGMENT_RECORD far * lpSegRec;
    HMODULE hModule;

    // Major hack!  We need an HMODULE here, not the module name.  We
    // could call GetModuleHandle(), but it's a relatively slow function
    // that will scan through all the module databases, doing strcmp()'s.
    // This will screw up our timings in some cases.  Instead, it just so
    // happens that the lpszModule given by TOOLHELP for NFY_LOADSEG
    // is a pointer to the module name in the module database.  Therefore,
    // the selector portion of this LPSTR is in fact the HMODULE.  If
    // this changes, we're hosed!.
    hModule = SELECTOROF(lpszModule);
    if ( *(LPWORD)MAKELP(hModule,0) != 0x454E ) // Verify it's an HMODULE
        return FALSE;                           // by looking for 'NE'

    lpSegRec = MAKELP( (WORD)hModule, 0x40 );

    return (BOOL) lpSegRec[segNumber-1].segRec.preload;
}

//
// Scan through the events, looking for the STARTTASK_EVENT and
// STARTDLL_EVENTS.  For each EXE or DLL, call HasDuplicateFixups()
// in DUPFIXUP.C to see how many duplicate fixups the file has.
//
void ScanForDuplicateFixups(void)
{
    char szModuleName[260];
    char buffer[384];
    WORD i, currentLastLine;
    HCURSOR oldCursor = 0;

    oldCursor = SetCursor( LoadCursor(0, IDC_WAIT) );

    currentLastLine = (WORD)SendMessage(HWndEventsLb, LB_GETCOUNT, 0, 0);

    for ( i=0; i < cLoadProfEvents; i++ )
    {
        DWORD cDuplicateFixups;
        ATOM filenameAtom;

        if ( LoadProfEvents[i].type == STARTDLL_EVENT )
            filenameAtom = LoadProfEvents[i].x.startDll;
        else if ( LoadProfEvents[i].type == STARTTASK_EVENT )
            filenameAtom = LoadProfEvents[i].x.startTask;
        else
            continue;

        GetAtomName( filenameAtom, szModuleName, sizeof(szModuleName) );

        cDuplicateFixups = HasDuplicateFixups(szModuleName);
        if ( cDuplicateFixups )
        {
            wsprintf(buffer, "%s has %lu duplicate fixups (%lu bytes)",
                (LPSTR)szModuleName, cDuplicateFixups,
                cDuplicateFixups * sizeof(COMMON_RELOCATION_RECORD));
            SendMessage(HWndEventsLb,LB_ADDSTRING,0, (LPARAM)(LPSTR)buffer);
        }
    }

    // Adjust the listbox so that it points at the newly added stuff
    SendMessage(HWndEventsLb, LB_SETTOPINDEX, currentLastLine, 0);

    SetCursor(oldCursor);
}

// Use COMMDLG.DLL to browse for the name of the EXE to profile
BOOL GetProgramName(HWND hWndOwner, LPSTR szFile, unsigned nFileBuffSize)
{
    OPENFILENAME ofn;
    static char szFilter1[] = "Programs (*.EXE)\0*.EXE\0";

    szFile[0] = 0;
    memset(&ofn, 0, sizeof(OPENFILENAME));
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hWndOwner;
    ofn.lpstrFilter = szFilter1;
    ofn.nFilterIndex = 1;
    ofn.lpstrFile= szFile;
    ofn.nMaxFile = nFileBuffSize;
    ofn.nMaxFileTitle = 0;
    ofn.lpstrFileTitle = (LPSTR)ofn.lpstrInitialDir = 0;
    ofn.lpstrTitle = "Program to profile";
    ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

    return GetOpenFileName(&ofn);
}

//
// Copy contents of listbox and time fields to a user specified file
//
void WriteReportToFile(HWND hWndOwner)
{
    OPENFILENAME ofn;
    char szFile[260];
    FILE *pOutFile;
    char buffer[260];
    WORD i, cLines;

    // use COMMDLG.DLL to browse for the name to save as
    memset(&ofn, 0, sizeof(OPENFILENAME));
    szFile[0] = '\0';
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hWndOwner;
    ofn.lpstrFile= szFile;
    ofn.nMaxFile = sizeof(szFile);
    if ( !GetSaveFileName(&ofn) )
        return;

    pOutFile = fopen((LPSTR)ofn.lpstrFile, "wt");
    if ( !pOutFile )
    {
        MessageBox(hWndOwner,"Couldn't open specified report file", 0, MB_OK);
        return;
    }

    GetDlgItemText(hWndOwner, IDC_STATIC_SEG_LOAD, buffer, sizeof(buffer));
    fprintf(pOutFile, "Seg Load: %s\n", buffer);
    GetDlgItemText(hWndOwner, IDC_STATIC_DLL_INIT, buffer, sizeof(buffer));
    fprintf(pOutFile, "DLL Init: %s\n", buffer);
    GetDlgItemText(hWndOwner, IDC_STATIC_APP_INIT, buffer, sizeof(buffer));
    fprintf(pOutFile, "App Init  %s\n", buffer);
    GetDlgItemText(hWndOwner, IDC_STATIC_TOTAL_TIME, buffer, sizeof(buffer));
    fprintf(pOutFile, "Total:    %s\n", buffer);

    cLines = (WORD)SendMessage(HWndEventsLb, LB_GETCOUNT, 0, 0);
    for ( i=0; i < cLines; i++ )
    {
        SendMessage(HWndEventsLb, LB_GETTEXT, i, (LPARAM)(LPSTR)buffer);
        fputs(buffer, pOutFile);
        fputs("\n", pOutFile);
    }
    fclose(pOutFile);
}

#ifdef __BORLANDC__
// Get rid of BC++ exception handling code - besides making the program
// fatter, it adds RC warnings about multi-instance fixup problems.
// See July 1994 Q&A for an explanation of this trick.
void _ExceptInit(void){}
#endif
